/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file server/tcpserver.c
 * @brief TCP server (gnunetd-client communication using util/tcpio.c).
 * @author Christian Grothoff
 **/

#include "gnunet_util.h"
#include "tcpserver.h"
#include "handler.h"

#define DEBUG_TCPSERVER NO

/**
 * Array of the message handlers.
 **/
static CSHandler * handlers = NULL;

/**
 * Number of handlers in the array (max, there
 * may be NULL pointers in it!)
 **/
static int max_registeredType = 0;

/**
 * Mutex to guard access to the handler array.
 **/
static Mutex handlerLock;

static int octets_total_tcp_in;

/**
 * The thread that waits for new connections.
 **/
static pthread_t TCPLISTENER_listener_;

/**
 * Handlers to call if client exits.
 **/
static ClientExitHandler * exitHandlers = NULL;
static int exitHandlerCount = 0;
static int exitHandlerSize = 0;

/**
 * Signals for control-thread to server-thread communication
 **/
static Semaphore * serverSignal = NULL;
static int tcpserver_keep_running = NO;

static Mutex lock;

typedef struct {
  pthread_t ct;
  int sock;
} ClientThreadHandle;

static ClientThreadHandle * socketList = NULL;
static int threadsLive = 0;

/**
 * Tell all applications that care that a client
 * has hung up.
 **/
static void cancelTCP(GNUNET_TCP_SOCKET * sock) {
  int i;

#if DEBUG_TCPSERVER
  LOG(LOG_DEBUG,
      "DEBUG: TCP Client exits, calling exit handlers\n");
#endif
  MUTEX_LOCK(&handlerLock);
  for (i=0;i<exitHandlerCount;i++) 
    exitHandlers[i](sock);  
  MUTEX_UNLOCK(&handlerLock);
}

int registerClientExitHandler(ClientExitHandler callback) {
  MUTEX_LOCK(&handlerLock);
  if (exitHandlerCount == exitHandlerSize) 
    GROW(exitHandlers,
	 exitHandlerSize,
	 exitHandlerSize + 10);
  exitHandlers[exitHandlerCount++] = callback;
  MUTEX_UNLOCK(&handlerLock);
  return OK;
}
  
int unregisterClientExitHandler(ClientExitHandler callback) {
  int i;

  MUTEX_LOCK(&handlerLock);
  for (i=0;i<exitHandlerSize;i++) {
    if (exitHandlers[i] == callback) {
      exitHandlers[i] = exitHandlers[--exitHandlerCount];
      MUTEX_UNLOCK(&handlerLock);
      return OK;
    }
  }
  MUTEX_UNLOCK(&handlerLock);
  return SYSERR;
}
  

/**
 * Handle a message (that was decrypted if needed).
 * Checks the CRC and if that's ok, processes the
 * message by calling the registered handler for
 * each message part.
 **/
static void processHelper(CS_HEADER * msg,
			  GNUNET_TCP_SOCKET * sender) {
  unsigned short ptyp;
  CSHandler callback;

  ptyp = htons(msg->tcpType);
  if (ptyp >= max_registeredType) {
    LOG(LOG_INFO, 
	"INFO: processHelper: Message not understood: %d (no handler registered, max is %d)\n",
	ptyp, 
	max_registeredType);
    return;
  }

  callback = handlers[ptyp];
  if (callback == NULL) 
    LOG(LOG_INFO, 
	"INFO: processHelper: Message not understood: %d (no handler registered)!\n",
	ptyp);
  else
    callback(sender,
	     msg);
}

/**
 * Handle data available on the TCP socket descriptor. This method
 * first aquires a slot to register this socket for the writeBack
 * method (@see writeBack) and then demultiplexes all TCP traffic
 * received to the appropriate handlers.  
 * @param sockDescriptor the socket that we are listening to (fresh)
 **/
static void * processData(int sockDescriptor){
  CS_HEADER * hdr;
  GNUNET_TCP_SOCKET sock;
  int i;

  hdr = MALLOC(MAX_BUFFER_SIZE);
  /* register the socket */
  initGNUnetServerSocket(sockDescriptor,
			 &sock);
  while (1) {
    if (SYSERR == readFromSocket(&sock,
				 (CS_HEADER**)&hdr) )
      break; /* connection closed */
    statChange(octets_total_tcp_in,
	       ntohs(hdr->size));
    processHelper(hdr,
		  &sock);
		  
  }
#if DEBUG_TCPSERVER
  LOG(LOG_DEBUG, 
      "DEBUG: closing TCP connection.\n");
#endif
  MUTEX_LOCK(&lock);
  cancelTCP(&sock);
  destroySocket(&sock);
  for (i=0;i<threadsLive;i++)
    if (socketList[i].sock == sockDescriptor)
      socketList[i].sock = -1;
  MUTEX_UNLOCK(&lock);
  FREE(hdr);
  return NULL;
}

/**
 * Initialize the TCP port and listen for incoming connections.
 **/
static void * tcpListenMain() {
  int i;
  int listenerFD;
  int incomingFD;
  int lenOfIncomingAddr;
  int listenerPort;
  struct sockaddr_in serverAddr, clientAddr;
  int secs = 5;
  const int on = 1;

  listenerPort = getGNUnetPort(); 
  /* create the socket */
 CREATE_SOCKET:
  while ( (listenerFD = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    LOG(LOG_ERROR, 
	"ERROR opening socket (%s). No client service started. Trying again in 30 seconds.\n",
	strerror(errno));
    sleep(30);
  }
  
  /* fill in the inet address structure */
  memset((char *) &serverAddr,
	 0,
	 sizeof(serverAddr));
  serverAddr.sin_family = AF_INET;
  serverAddr.sin_addr.s_addr=htonl(INADDR_ANY);
  serverAddr.sin_port=htons(listenerPort);
 
  if ( setsockopt(listenerFD, 
		  SOL_SOCKET, 
		  SO_REUSEADDR, 
		  &on, sizeof(on)) < 0 )
    perror("setsockopt");

  /* bind the socket */
  if (bind (listenerFD, 
	   (struct sockaddr *) &serverAddr,
	   sizeof(serverAddr)) < 0) {
    LOG(LOG_ERROR, 
	"ERROR (%s) binding the TCP listener to port %d. No proxy service started.\nTrying again in %d seconds...\n",
	strerror(errno),
	listenerPort, 
	secs);
    sleep(secs);
    secs += 5; /* slow progression... */
    CLOSE(listenerFD);
    goto CREATE_SOCKET;
  }
  
  /* start listening for new connections */
  listen(listenerFD, 5); /* max: 5 pending, unhandled connections */
  SEMAPHORE_UP(serverSignal); 

  /* process incoming data */
  while (tcpserver_keep_running == YES) {
    /* wait for a connection and process it */
    lenOfIncomingAddr = sizeof(clientAddr);
    incomingFD = accept(listenerFD,
			(struct sockaddr *)&clientAddr, 
			&lenOfIncomingAddr);
    if(incomingFD < 0) {
      if ( (tcpserver_keep_running == YES) &&
	   (errno != EINTR) )
	LOG(LOG_ERROR, 
	    "ERROR accepting new connection (%s).\n",
	    strerror(errno));
      continue;
    }
    /* verify clientAddr for eligibility here (ipcheck-style,
       user should be able to specify who is allowed to connect,
       otherwise we just close and reject the communication! */  
    if (NO == isWhitelisted(clientAddr.sin_addr)) {
      LOG(LOG_WARNING,
	  "WARNING: Rejected unauthorized connection from %d.%d.%d.%d.\n",
	  PRIP(ntohl(*(int*)&clientAddr.sin_addr)));
      CLOSE(incomingFD);
      continue;
    }

    /* printConnectionBuffer();*/
#if DEBUG_TCPSERVER
    LOG(LOG_DEBUG, 
	"DEBUG: starting tcp server!\n");
#endif
    MUTEX_LOCK(&lock);
    /* test if another thread has terminated and we
       can recycle the slot. If not, create a new one */
    for (i=0;i<threadsLive;i++)
      if (socketList[i].sock == -1) {
	void * unused;
	PTHREAD_JOIN(socketList[i].ct, &unused);
	break;
      }
    if (i == threadsLive) {
      i = threadsLive;
      GROW(socketList,
	   threadsLive,
	   threadsLive+1);
    }
    socketList[i].sock = incomingFD;

    /* create the thread */
    if ((PTHREAD_CREATE(&socketList[i].ct,
			(PThreadMain) &processData, 
			(void *) incomingFD,
			1024 * 256)) != 0) {
      LOG(LOG_ERROR, 
	  "ERROR creating thread to handle new incoming connection.\n");    
      CLOSE(incomingFD);
      /* shrink! */
      socketList[i] = socketList[threadsLive-1];
      GROW(socketList,
	   threadsLive,
	   threadsLive-1);
    } 
    MUTEX_UNLOCK(&lock);
  } /* while (keep_running) */
  SEMAPHORE_UP(serverSignal); 
  return NULL;
} 

/**
 * Initialize the TCP port and listen for incoming client connections.
 **/
int initTCPServer() {
  MUTEX_CREATE(&lock);
  if (tcpserver_keep_running == YES) {
    LOG(LOG_FATAL,
	"FATAL: initTCPServer called, but TCPserver is already running\n");
    return SYSERR;
  }
  octets_total_tcp_in 
    = statHandle("# bytes received from clients");
  MUTEX_CREATE(&handlerLock);
  tcpserver_keep_running = YES;
  serverSignal = SEMAPHORE_NEW(0);
  if (0 == PTHREAD_CREATE(&TCPLISTENER_listener_, 
			  (PThreadMain) &tcpListenMain, 
			  NULL,
			  512)) {
    SEMAPHORE_DOWN(serverSignal);
  } else {
    LOG(LOG_FAILURE, 
	"FAILURE: could not start TCP server (pthread error!?)\n");
    SEMAPHORE_FREE(serverSignal);
    serverSignal = NULL;
    tcpserver_keep_running = NO;
    MUTEX_DESTROY(&handlerLock);
    return SYSERR;
  }
  /* register default handlers */
  registerCSHandler(STATS_CS_PROTO_GET_STATISTICS,
		    &sendStatistics);
  registerCSHandler(STATS_CS_PROTO_GET_CS_MESSAGE_SUPPORTED,
		    &handleCSMessageSupported);
  registerCSHandler(STATS_CS_PROTO_GET_P2P_MESSAGE_SUPPORTED,
		    &handlep2pMessageSupported);
  return OK;
}

/**
 * Shutdown the module.
 **/ 
int stopTCPServer() {
  if ( ( tcpserver_keep_running == YES) &&
       ( serverSignal != NULL) ) {
    GNUNET_TCP_SOCKET sock;
    IPaddr ip;
    int i;
    void * unused;

    /* stop server thread */
    tcpserver_keep_running = NO;
    *(int*)&ip = htonl(0x7F000001); /* 127.0.0.1 = localhost */ 
    initGNUnetClientSocketIP(getGNUnetPort(),
			     ip,
			     &sock);
    if (SYSERR == checkSocket(&sock)) 
      errexit("FATAL: could not contact gnunetd server for shutdown\n");
    destroySocket(&sock);
    SEMAPHORE_DOWN(serverSignal);
    SEMAPHORE_FREE(serverSignal);  
    PTHREAD_JOIN(TCPLISTENER_listener_, &unused);

    /* stop client threads */
    for (i=0;i<threadsLive;i++) {
      if (socketList[i].sock != -1)
	shutdown(socketList[i].sock, 2);
      PTHREAD_JOIN(socketList[i].ct, &unused);
    }      
    GROW(socketList,
	 threadsLive,
	 0);
    return OK;
  } else
    return SYSERR;
}

int doneTCPServer() {
  /* free data structures */
  MUTEX_DESTROY(&handlerLock);
  GROW(handlers,
       max_registeredType,
       0);
  MUTEX_DESTROY(&lock);
  GROW(exitHandlers,
       exitHandlerSize,
       0);
  return OK;
}

/**
 * Register a method as a handler for specific message
 * types. 
 * @param type the message type
 * @param callback the method to call if a message of
 *        that type is received, if the callback returns
 *        SYSERR, processing of the message is discontinued
 *        afterwards (all other parts are ignored)
 * @return OK on success, SYSERR if there is already a
 *         handler for that type
 **/
int registerCSHandler(const unsigned short type,
		      CSHandler callback) {
  MUTEX_LOCK(&handlerLock);
  if (type < max_registeredType) {
    if (handlers[type] != NULL) {
      MUTEX_UNLOCK(&handlerLock);
      LOG(LOG_WARNING,
	  "WARNING: registerCSHandler failed, slot %d used\n",
	  type);
      return SYSERR;
    } 
  } else
    GROW(handlers,
	 max_registeredType,
	 type + 32);
  handlers[type] = callback;
  MUTEX_UNLOCK(&handlerLock);
  return OK; 
}

/**
 * Return wheter or not there is a method handler 
 * registered for a specific Client-Server message type.
 *
 * @param type the message type
 * @return YES if there is a handler for the type,
 * 	NO if there isn't
 **/
int isCSHandlerRegistered(const unsigned short type) {
    int registered = NO;

    if ((type < max_registeredType) &&
    	(handlers[type] != NULL)) {
        registered = YES;
    }

    return registered;
}

  
/**
 * Unregister a method as a handler for specific message
 * types. 
 *
 * @param type the message type
 * @param callback the method to call if a message of
 *        that type is received, if the callback returns
 *        SYSERR, processing of the message is discontinued
 *        afterwards (all other parts are ignored)
 * @return OK on success, SYSERR if there is no or another
 *         handler for that type
 **/
int unregisterCSHandler(const unsigned short type,
			CSHandler callback) {
  MUTEX_LOCK(&handlerLock);
  if (type < max_registeredType) {
    if (handlers[type] != callback) {
      MUTEX_UNLOCK(&handlerLock);
      return SYSERR; /* another handler present */
    } else {
      handlers[type] = NULL;
      MUTEX_UNLOCK(&handlerLock);
      return OK; /* success */
    }
  } else {  /* can't be there */
    MUTEX_UNLOCK(&handlerLock);
    return SYSERR;
  }
}


/**
 * Handle a request to see if a particular client server message 
 * is supported.
 **/
int handleCSMessageSupported(GNUNET_TCP_SOCKET * sock,
		   CS_HEADER * message) {
  unsigned short type;
  int supported;
  STATS_CS_GET_MESSAGE_SUPPORTED * cmsg;

  if (ntohs(message->size) != sizeof(STATS_CS_GET_MESSAGE_SUPPORTED)) {
    LOG(LOG_WARNING,
	"WARNING: message received from client is invalid.\n");
    return SYSERR;
  }
  cmsg = (STATS_CS_GET_MESSAGE_SUPPORTED *) message;
  type = ntohs(cmsg->tcpType);

  supported = isCSHandlerRegistered( type );

  return sendTCPResult(sock, supported);
}

/* end of tcpserver.c */
